/*
	Copyright (C) 2006,2007 Hewlett-Packard Development Company, L.P.
	Author: Stephen M. Cameron

	This file is part of cciss_vol_status.

	cciss_vol_status is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	cciss_vol_status is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with cciss_vol_status; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#define VERSION_NUMBER PACKAGE_VERSION 

#include <stdio.h>

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#elif HAVE_STRINGS_H
#include <strings.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#if HAVE_ERRNO_H
#include <errno.h>
#else
extern int errno; /* some systems #define this! */
#endif /* HAVE_ERRNO_H */

#include <libgen.h>
#include <sys/ioctl.h>
#include <inttypes.h>
#ifdef HAVE_SCSI_SG_H
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <netinet/in.h> /* for htons() */
#endif

#ifdef HAVE_LINUX_CCISS_IOCTL_H
/* Some versions of cciss_ioctl.h contain a "__user" attribute which 
   needs masking out to compile outside the kernel. */
#define __user
#include <linux/cciss_ioctl.h>
#elif HAVE_CISSIO_H
/* FreeBSD has cissio.h as a cciss_ioctl.h doppelganger */
#include "/usr/src/sys/dev/ciss/cissio.h"
#else
#error Neither cciss_ioctl.h (linux) nor cissio.h (FreeBSD) were found.
#endif

#define ID_CTLR 0x11 /* Identify controller */
#define ID_LSTATUS 0x12 /* identify logical drive status */

int everything_hunky_dory = 1;
int be_quiet = 0;
int exhaustive_search = 0;

#define MSA1000_ID 0xe0100e11

#ifndef SCSI_IOCTL_SEND_COMMAND
#define SCSI_IOCTL_SEND_COMMAND 1
#endif

struct smartarray_id_t {
	uint32_t board_id;
	char *board_name;
	int supported_by_this_program;
} smartarray_id[] = {
	/* this table heavily Borrowed from linux kernel,
	   from drivers/block/cciss.c */
 	{ 0x40700E11, "Smart Array 5300",	1 },
 	{ 0x40800E11, "Smart Array 5i",		1,},
 	{ 0x40820E11, "Smart Array 532",	1,},
 	{ 0x40830E11, "Smart Array 5312",	1,},
 	{ 0x409A0E11, "Smart Array 641",	1,},
 	{ 0x409B0E11, "Smart Array 642",	1,},
 	{ 0x409C0E11, "Smart Array 6400",	1,},
 	{ 0x409D0E11, "Smart Array 6400 EM",	1,},
 	{ 0x40910E11, "Smart Array 6i",		1,},
 	{ 0x409E0E11, "Smart Array 6422",	1,},
 	{ 0x3225103C, "Smart Array P600",	0,},
	{ 0x3234103C, "Smart Array P400",	0,},
	{ 0x3235103C, "Smart Array P400i",	0,},
	{ 0x3211103C, "Smart Array E200i",	0,},
	{ 0x3212103C, "Smart Array E200",	0,},
	{ 0x3213103C, "Smart Array E200i",	0,},
	{ 0x3214103C, "Smart Array E200i",	0,},
	{ 0x3215103C, "Smart Array E200i",	0,},
	{ 0x3223103C, "Smart Array P800",	0,},
#ifdef HAVE_SCSI_SG_H
	{ MSA1000_ID, "MSA1000",	1,},
#else
#warning Since <scsi/sg.h> is not around, MSA1000 support will not be compiled.
#endif
	{ 0xFFFFFFFF, "Unknown Smart Array",	0,},
};

/* See the following documents for information about the hardware 
 * specific structures used in this program:
 *
 *	fwspecwww.doc from http://sourceforge.net/progjects/cpqarray
 *	Open_CISS_Spec.pdf from http://sourceforge.net/progjects/cciss
 *	and see also the ciss.h header file from the freebsd ciss driver
 */

/* Structure returned by command to get logical drive status (0x12) */
#pragma pack(1)
typedef struct
{
	unsigned char status;
	uint32_t drive_failure_map;
	unsigned char reserved[416];
	uint32_t blocks_left_to_recover;
	unsigned char drive_rebuilding;
	uint16_t remap_count[32];
	uint32_t replacement_drive_map;
	uint32_t active_spare_map;
	unsigned char spare_status;
	unsigned char spare_to_replace_map[32];
	uint32_t replaced_marked_ok_map;
	unsigned char media_exchanged;
	unsigned char cache_failure;
	unsigned char expand_failure;
	unsigned char unit_flags;
	uint16_t big_failure_map[8];
	uint16_t big_remap_cnt[128];
	uint16_t big_replace_map[8];
	uint16_t big_spare_map[8];
	unsigned char  big_spare_replace_map[128];
	uint16_t big_replace_ok_map[8];	
	unsigned char  big_drive_rebuild;
} ida_id_lstatus;
#pragma pack()

const char *spare_drive_status_msg[] = {
		/* Corresponds to bits in spare_status field, above */
		/* bit 0 */  "At least one spare drive",
		/* bit 1 */  "At least one spare drive activated and currently rebuilding",
		/* bit 2 */  "At least one spare drive has failed",
		/* bit 3 */  "At least one spare drive activated",
		/* bit 4 */  "At least one spare drive remains available",
};
#define NSPARE_MSGS (sizeof(spare_drive_status_msg) / sizeof(spare_drive_status_msg[0]))

#pragma pack(1)
/* Structure returned by Identify Controller command (0x11) */
typedef struct {
	unsigned char num_logical_drives;
	uint32_t signature;
	uint32_t running_firm_rev;
	unsigned char rom_firm_rev[4];
	unsigned char hardware_rev;
	unsigned char reserved[4];
	uint32_t drive_present_bit_map;
	uint32_t external_drive_bit_map;
	uint32_t board_id;
	unsigned char reserved2;
	uint32_t non_disk_map;
	unsigned char reserved3[5]; 
	unsigned char marketing_revision;
	unsigned char controller_flags;
	unsigned char reserved4[2];
	unsigned char expand_disable_code;
	unsigned char scsi_chip_count;
	uint32_t reserved5;
	uint32_t ctlr_clock;
	unsigned char drives_per_scsi_bus;
	uint16_t big_drive_present_map[8];
	uint16_t big_ext_drive_map[8]; 
	uint16_t big_non_disk_map[8];
} id_ctlr;
#pragma pack()

char *progname = "cciss_vol_status";

void usage()
{
	fprintf(stderr, "%s: usage: %s [ -q ] [ -v ] /dev/cciss/c*d0\n", progname, progname);
	fprintf(stderr, "	-q	Be quiet.\n");
	fprintf(stderr, "	-v	Print program version and exit.\n");
	fprintf(stderr, "	-x	exhaustive search for logical drives.\n");
	exit(-1);
}

int bitisset(unsigned char bitstring[], int bit, int bitstringlength)
{
	int element;
	int offset;
	unsigned char or_val;

	element = (bit / 8);
	offset  = bit % 8;
	or_val = (unsigned char) (1 << offset);

	return ((bitstring[element] & or_val) != 0);
}

int id_ctlr_fd(char * filename, int fd, id_ctlr *id)
{
	int rc;
	IOCTL_Command_struct c;

	memset(id, 0, sizeof(*id));
	memset(&c, 0, sizeof(c));
	c.Request.CDBLen = 10;
	c.Request.Type.Type = TYPE_CMD;
	c.Request.Type.Attribute = ATTR_SIMPLE;
	c.Request.Type.Direction = XFER_READ;
	c.Request.Timeout = 0;
	c.Request.CDB[0] = 0x26;
	c.Request.CDB[6] = ID_CTLR;
	c.buf_size = sizeof(*id);
	c.Request.CDB[7] = (c.buf_size >> 8) & 0xff;
	c.Request.CDB[8] = c.buf_size  & 0xff;
	c.buf = (unsigned char *) id;
	rc = ioctl(fd, CCISS_PASSTHRU, &c);
	if (rc != 0) {
		fprintf(stderr, "%s: Cannot identify controller %s, ioctl failed: %s\n",
			progname, filename, strerror(errno));
		return -1;
	}

	if (c.error_info.CommandStatus != 0) { /* Shouldn't happen */
		fprintf(stderr, "Identify controller: Commandstatus is %d\n", c.error_info.CommandStatus);
		return -1;
	}

	if (id->drives_per_scsi_bus == 0)  /* not just CASA apparently. */
		id->drives_per_scsi_bus = 7;
	return(0);
}

void show_disk_map(char *title,
		id_ctlr *id,
		unsigned char bigmap[],
		uint32_t smallmap
		)
{
	int bus, target, lastbus;
	int i, count;
	int big_map_support;
	int ndisks=0;
	int drives_per_scsi_bus;

	lastbus = -1;

	big_map_support = (id->controller_flags && (1<<7));
	drives_per_scsi_bus = id->drives_per_scsi_bus;;

	/* if no big drive map support, & zero devices per scsi bus */
	if (!big_map_support && (drives_per_scsi_bus == 0))
		drives_per_scsi_bus = 7;

	if (drives_per_scsi_bus == 0) {
		/* should never happen, but guard against. */
		fprintf(stderr, "Controller reports zero devices per "
			"scsi bus.  This is not reasonable. Exiting.\n");
		exit(-1);
	}

	for (i=0;i < (big_map_support ? 128 : 16);i++)  { /* FIXME magic 128,16 */
		bus    = i / drives_per_scsi_bus;
		target = i % drives_per_scsi_bus;

	  	if ( (big_map_support &&
			bitisset((unsigned char *) bigmap, i, 128 / 8)) ||
			( (! big_map_support) &&
			bitisset((unsigned char *) &smallmap, i, sizeof(smallmap)))) {
			if (lastbus == -1)
				printf("%s", title);
			if (bus != lastbus) {
				lastbus = bus;
				count=0;
				printf("\n// bus %d:\n//    ", bus);
			}

			if (count != 0)
				printf(", ");

			lastbus = bus;
			printf("b%dt%d", bus, target);
			count++; ndisks++;

			if (count > 8)  {  /* reasonable for 80 column display */
				count=0;
				printf("\n//    ");
			}
		}
	}
	if (lastbus != -1)
		printf("\n");
}

char *decode_status[] = {
	/*  0 */ "OK",
	/*  1 */ "FAILED",
	/*  2 */ "Not configured",
	/*  3 */ "Using interim recovery mode",
	/*  4 */ "Ready for recovery operation",
	/*  5 */ "Currently recovering",

	/* Note: If the unit_status value is 6 or 7, the unit_status 
		of all other configured logical drives will be marked as 
		1 (Logical drive failed). This is to force the user to 
		correct the problem and to insure that once the problem 
		is corrected, the data will not have been corrupted by 
		any user action. */

	/*  6 */ "Wrong physical drive was replaced",
	/*  7 */ "A physical drive is not properly connected",

	/*  8 */ "Hardware is overheating",
	/*  9 */ "Hardware was overheated",
	/* 10 */ "Currently expannding",
	/* 11 */ "Not yet available",
	/* 12 */ "Queued for expansion",
};

#define MAX_DECODE_STATUS (sizeof(decode_status) / sizeof(decode_status[0]))

void print_volume_status(char *file, int ctlrtype, int volume_number, 
	ida_id_lstatus *vs, id_ctlr *id)
{
	int spare_bit;

	if (vs->status == 2) /* This logical drive is not configured, so no status. */
		return;

	if (vs->status != 0)
		everything_hunky_dory = 0;

	printf("%s: (%s) Volume %d status: ", file, 
		smartarray_id[ctlrtype].board_name, volume_number);
	if (vs->status < MAX_DECODE_STATUS)
		printf("%s. ", decode_status[vs->status]);
	else
		printf("Unknown status %d.", vs->status);
	
	for (spare_bit=0; spare_bit < NSPARE_MSGS; spare_bit++)
		if ((vs->spare_status >> spare_bit) & 0x01)
			printf("  %s.", spare_drive_status_msg[spare_bit]);

	if (vs->status != 0 && !smartarray_id[ctlrtype].supported_by_this_program)
		printf("%s: Don't know how to decode drive map data for %s controllers.\n",
			progname, smartarray_id[ctlrtype].board_name);

	if (smartarray_id[ctlrtype].supported_by_this_program) {
		show_disk_map("  Failed drives:", id,
				(unsigned char *) vs->big_failure_map, vs->drive_failure_map);
		show_disk_map("  'Replacement' drives:", id,
				(unsigned char *) vs->big_replace_map, vs->replacement_drive_map);
		show_disk_map("  Drives currently substituted for by spares:", id,
				(unsigned char *) vs->big_spare_map, vs->active_spare_map);
	}

	printf("\n");
}

int lookup_controller(uint32_t board_id)
{
	int i;

	for (i=0;smartarray_id[i].board_id != 0xFFFFFFFF; i++)
		if (smartarray_id[i].board_id == board_id)
			return i;
	return -1;
}


#ifdef HAVE_SCSI_SG_H

void setup_sgio(sg_io_hdr_t *sgio,
		unsigned char *cdb, unsigned char cdblen,
		unsigned char *databuffer, unsigned int databufferlen, 
		unsigned char *sensebuffer, unsigned char sensebufferlen,
		int direction
	       	)
{
	sgio->interface_id = 'S';
	sgio->dxfer_direction = direction;
	sgio->cmd_len = cdblen;
	sgio->mx_sb_len = sensebufferlen;
	sgio->iovec_count = 0;
	sgio->dxfer_len = databufferlen;
	sgio->dxferp = databuffer;
	sgio->cmdp = cdb;
	sgio->sbp = sensebuffer;
	sgio->timeout = 0x0fffffff; /* long timeout */
	/* I used a long timeout here instead of no timeout (0xffffffff) because */
	/* some drivers (mptlinux) appear not to like a timeout value of */
	/* 0xffffffff at all.  This way, using cciss_vol_status indiscriminately */
	/* (e.g. cciss_vol_status /dev/sg* ) doesn't make mptlinux driver mad. */

	sgio->flags = 0;
	sgio->pack_id = 0;
	sgio->usr_ptr = NULL;
	sgio->status = 0;
	sgio->masked_status = 0;
	sgio->msg_status = 0;
	sgio->sb_len_wr = 0;
	sgio->host_status = 0;
	sgio->driver_status = 0;
	sgio->resid = 0;
	sgio->duration = 0;
	sgio->info = 0;
}

#define debug_sgio 0

static inline int min(int x, int y)
{
	return x < y ? x : y;
}

int do_sg_io(int fd, unsigned char *cdb, unsigned char cdblen, unsigned char *buffer,
	unsigned int buf_size, int direction)
{
	int status;
	sg_io_hdr_t sgio;
	unsigned char sensebuffer[64];

	memset(buffer, 0, buf_size);
	memset(&sgio, 0, sizeof(sgio));

	setup_sgio(&sgio, (unsigned char *) cdb, cdblen, buffer,
		buf_size, sensebuffer, sizeof(sensebuffer), direction);

	status = ioctl(fd, SG_IO, &sgio);

	if (status == 0 && sgio.host_status == 0 && sgio.driver_status == 0) {   /* cmd succeeded */
		if (sgio.status == 0 || (sgio.status == 2 &&
			(((sensebuffer[2] & 0x0f) == 0x00) || /* no error */
			((sensebuffer[2] & 0x0f) == 0x01)))) {
			return 0;
		}
		if (debug_sgio)
			fprintf(stderr, "sgio cmd 0x%02x check condition, sense key = %d\n",
				cdb[0], sensebuffer[2] & 0x0f);
		return -1;
	}
	if (debug_sgio)
		fprintf(stderr, "sgio ioctl: %d, cdb[0]=0x%02x, "
			"status host/driver/scsi/sensekey = %d/%d/%d/0x%02x\n",
			status, cdb[0], sgio.host_status, sgio.driver_status,
			sgio.status, sensebuffer[2]);
	return -1;
}

int do_inquiry(int fd,
               unsigned char inquiry_page,
               unsigned char* inquiry_buf,
               unsigned char buf_size)
{
	unsigned char cdb[6];

	memset(inquiry_buf, 0, buf_size);
	memset(cdb, 0, sizeof(cdb));
	cdb[0] = 0x12;
	if (inquiry_page != 0) /* EVPD? */
		cdb[1] = 1;
	else
		cdb[1] = 0;
	cdb[2] = inquiry_page;
	cdb[4] = buf_size;

	return do_sg_io(fd, cdb, 6, inquiry_buf, (unsigned int) buf_size, SG_DXFER_FROM_DEV);
}

int msa1000_passthru_ioctl (int fd, int cmd, void *buffer, int size, uint log_unit)
{ 
	sg_io_hdr_t sgio, sgio2;
	unsigned char sensebuffer[64];
	int direction = SG_DXFER_FROM_DEV;
	unsigned char cdb[16];
	int cdblen, nbytes;
	unsigned short bufsize;

	memset(cdb, 0, sizeof(cdb));
	memset(sensebuffer, 0, 64);
	switch (cmd) {
	case 0x10: /* ID_LDRIVE: */
	case ID_LSTATUS: {
	        bufsize = htons((unsigned short) size);
	        cdb[0] = 0x26;
		cdb[1] = log_unit;
		cdb[6] = cmd;
		memcpy(&cdb[7], &bufsize, sizeof(bufsize));
		direction = SG_DXFER_FROM_DEV;
		cdblen = 10;
		break;
		}
	case ID_CTLR: {	/* 0x11 */ 
	        bufsize = htons((unsigned short) size);
	        cdb[0] = 0x26;
		cdb[6] = cmd;
		memcpy(&cdb[7], &bufsize, sizeof(bufsize));
		direction = SG_DXFER_FROM_DEV;
		cdblen = 10;
		break;
		}
	default:
		fprintf(stderr, "msa1000_passthru_ioctl: unknown command %d\n", cmd);
		return -1;
	}
	return do_sg_io(fd, cdb, cdblen, buffer, size, direction);
}

int is_msa1000(int fd)
{
	int status;
	char std_inq[256];
	char *prod =  "COMPAQ  MSA1000";
	char *prod2 =  "MSA CONTROLLER";

	status = do_inquiry(fd, 0, (unsigned char *) std_inq, 255);
	if (status < 0) 
		return 0;
	if (strncmp(std_inq+8, prod, strlen(prod)) == 0)
		return 1;
	if (strstr(std_inq+8, prod2) != 0)
		return 1;
	return 0;
}

void msa1000_logical_drive_status(char *file, int fd, 
	unsigned int logical_drive, id_ctlr *id)
{
	int rc, ctlrtype;
	ida_id_lstatus ldstatus;

	rc = msa1000_passthru_ioctl(fd, ID_LSTATUS, &ldstatus, 
		sizeof(ldstatus), logical_drive);
	if (rc < 0) {
		fprintf(stderr, "%s: %s: ioctl: logical drive: %d %s\n",
				progname, file, logical_drive, strerror(errno));
		return;
	}
	ctlrtype = lookup_controller(id->board_id);
	if (ctlrtype == -1) {
		fprintf(stderr, "%s: %s: Unknown controller, board_id = 0x%08x\n", 
			progname, file, id->board_id);
		return;
	}
	print_volume_status(file, ctlrtype, logical_drive, &ldstatus, id);
}

#endif /* ifdef HAVE_SCSI_SG_H */

int cciss_status(char *file)
{
	int fd, status, i, rc;
	IOCTL_Command_struct c;
	ida_id_lstatus ldstatus;
	id_ctlr id;
	struct stat statbuf;
	int ctlrtype;
	int numluns;
	int bus;

	fd = open(file, O_RDWR);
	if (fd < 0) {
		if (errno != ENXIO)
			fprintf(stderr, "%s: open %s: %s\n",
				progname, file, strerror(errno));
		return -1;
	}

#ifdef HAVE_SCSI_SG_H
	/* See if it's a scsi device... */
        rc = ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &bus);
        if (rc == 0) {
		/* It's a SCSI device... maybe it's an MSA1000. */
		if (is_msa1000(fd)) {
			rc = msa1000_passthru_ioctl(fd, ID_CTLR, &id, sizeof(id), 0);
			if (rc < 0) {
				fprintf(stderr, "%s: %s: Can't identify controller, id = 0x%08x.\n",
					progname, file, id.board_id);
				return -1;
			}
			numluns = (exhaustive_search) ? 32 : id.num_logical_drives;
			for (i=0;i<numluns;i++) /* We know msa1000 supports only 32 logical drives */
				msa1000_logical_drive_status(file, fd, i, &id);
			close(fd);
			return 0;
		}
		fprintf(stderr, "%s: %s: Unknown SCSI device.\n", progname, file);
		close(fd);
		return -1;
	}
#endif

	/* Stat the file to see if it's a block device file */
	rc = fstat(fd, &statbuf);
	if (rc < 0) {
		fprintf(stderr, "%s: cannot stat %s: %s\n",
			progname, file, strerror(errno));
		return -1;
	}

	/* On linux, we want a block device; on freebsd, character */
#if HAVE_CISSIO_H
#define WANTED_DEVICE_TYPE
#define WANTED_DEVICE_TYPE_STRING "character" 
#else
#define WANTED_DEVICE_TYPE !
#define WANTED_DEVICE_TYPE_STRING "block" 
#endif
	if (WANTED_DEVICE_TYPE S_ISBLK(statbuf.st_mode)) {
		fprintf(stderr, "%s: %s is not a %s device.\n",
			progname, file, WANTED_DEVICE_TYPE_STRING);
		return -1;
	}

	/* Not a SCSI device, see if it's a smartarray of some kind... */
	rc = id_ctlr_fd(file, fd, &id);
	if (rc < 0)
		return -1;

	/* Look it up in the table */
	ctlrtype = lookup_controller(id.board_id);
	if (ctlrtype == -1) {
		fprintf(stderr, "%s: %s: Unknown controller, board_id = 0x%08x\n", 
			progname, file, id.board_id);
		return -1;
	}

	/* Seems to be a cciss controller, Find the status of each logical drive */
	numluns = (exhaustive_search) ? 255 : id.num_logical_drives;
	for (i=0;i<numluns;i++) {
		/* Construct command to get logical drive status */
		memset(&c, 0, sizeof(c));
		c.LUN_info.LogDev.Mode = 0x0;
		c.Request.CDBLen = 10;
		c.Request.Type.Type = TYPE_CMD;
		c.Request.Type.Attribute = ATTR_SIMPLE;
		c.Request.Type.Direction = XFER_READ;
		c.Request.Timeout = 0;
		c.Request.CDB[0] = 0x26; /* 0x26 means "CCISS READ" */
		c.Request.CDB[1] = i; /* logical drive #  start from 0*/
		c.Request.CDB[6] = ID_LSTATUS;
		c.buf_size = sizeof(ldstatus);
		c.Request.CDB[7] = (c.buf_size >> 8) & 0xff;
		c.Request.CDB[8] = c.buf_size  & 0xff;
		c.buf = (unsigned char *) &ldstatus;

		/* Get logical drive status */
		status = ioctl(fd, CCISS_PASSTHRU, &c);
		if (status != 0) {
			fprintf(stderr, "%s: %s: ioctl: logical drive: %d %s\n",
				progname, file, i, strerror(errno));
			continue;
		}
		if (c.error_info.CommandStatus != 0) {
#ifndef DEBUG
			/* We ask for logical drive status for drives 0..255 */ 
			/* just to be sure to get all of them.  For those beyond */
			/* the controller's ability, (beyond 31, for older ctlrs) */
			/* we'll get back CommandStatus == 4 (invalid cmd) */
			if (c.error_info.CommandStatus == 4)
				continue;
#endif
			fprintf(stderr, "Error getting status for %s "
				"volume %d: Commandstatus is %d\n", 
				file, i, c.error_info.CommandStatus);
			continue;
		}
		print_volume_status(file, ctlrtype, i, &ldstatus, &id);
	}
	close(fd);
	return 0;
}

void intro()
{
	if (isatty(2) && !be_quiet) { /* Only print this message if stderr is a tty */
		fprintf(stderr, "%s version %s,\n"
			"Copyright (C) 2007 Hewlett-Packard Development Company, L.P.\n",
			progname, VERSION_NUMBER);
		fprintf(stderr, "%s comes with ABSOLUTELY NO WARRANTY.\n", progname);
		fprintf(stderr, "This is free software, and you are welcome to redistribute it\n");
		fprintf(stderr, "under certain conditions.  ");
		fprintf(stderr, "See the file 'COPYING', which you \n");
		fprintf(stderr, "should have received with this program, for details regarding\n");
		fprintf(stderr, "the absence of warranty and the conditions for redistribution.\n\n");
	}
}

int main(int argc, char *argv[])
{
	int i, opt;

	do {
		opt = getopt(argc, argv, "qvx");
		switch (opt) {
			case 'q': be_quiet = 1;
				continue;
			case 'v': fprintf(stderr, "%s version %s%s\n", progname, 
					VERSION_NUMBER, 
#ifdef HAVE_SCSI_SG_H
					""
#else
					"(without support for MSA1000)"
#endif
					);
				  exit(0);
			case 'x' : exhaustive_search = 1;
				continue;
			case '?' : 
			case ':' :
				usage(); /* usage calls exit(), so no fall thru */
			case -1 : 
				break;
		}
	} while (opt != -1);

	if ((argc - optind) < 1)
		usage();

	intro();

	for (i=optind;i<argc;i++)
		cciss_status(argv[i]);
	exit((everything_hunky_dory != 1));
}

